import time
import numpy as np
import scipy.sparse as sp
import matplotlib.pyplot as plt
import netgen.gui
from ngsolve import *
# from netgen.occ import *
from netgen.read_gmsh import ReadGmsh

plt.close('all')

ngsglobals.msg_level = 5

#%% import 3D geometry
#geo = OCCGeometry("Geometry/battery3d.step")
#mesh = Mesh(geo.GenerateMesh(maxh=5))
mesh = ReadGmsh("./Stirling_engine.msh")
mesh = Mesh(mesh)

mesh.ngmesh.SetBCName(19,"heating_boundary") #19#read out from .geo file
mesh.ngmesh.SetMaterial (1, "Plastic")
mesh.ngmesh.SetMaterial (2, "Air") 
mesh.ngmesh.SetMaterial (3, "Insert")
mesh.ngmesh.SetMaterial (4, "Frame")

print(mesh.GetBoundaries())
print(mesh.GetMaterials())

#%% 
#Temperature at Dirichlet boundary:
dirichlet_temperature_const = 60

#Define range of densities in the regions   [kg/m^3]
region_densities = np.array([1360,1225,8730,7850]) #Frame

#Define range of specific heat capacities in the regions  [J/K kg]
region_hcs = np.array([1170,1004,377,480]) 

#Define range of thermal conductivities in the regions [J/msK]
region_conductivities = np.array([0.19,0.025,150,46])

#Define range of specific heat capacities in the regions
region_losses = np.array([0,0,0,0])

    
#Initial value of solution
initial_value = 20 #[°C]
#compute snapshot diffusivities [m^2/s]
# region_diffusivities = region_conductivities / (region_densities * region_hcs)
region_diffusivities = region_conductivities
region_diffusivities = region_diffusivities * 1000000 # scale to [mm^2/s]

region_sourceterms = region_losses #  / (region_densities * region_hcs)

#%% define finite Element space

fes = H1(mesh, order=2, dirichlet="heating_boundary")
system_dofs = fes.ndof  # number of unknowns in this space

#%%
u = fes.TrialFunction()  # symbolic object
v = fes.TestFunction()   # symbolic object
gfu = GridFunction(fes)  # solution

# define thermal diffusivities
K = CoefficientFunction(list(region_diffusivities)) 

# define sourceterms
sources = CoefficientFunction(list(region_sourceterms)) 

#Draw(sources,mesh,'piecewise',draw_surf=False)
#%% set up the equation system

a = BilinearForm(fes, symmetric=False)
a += SymbolicBFI(K*grad(u)*grad(v)) 
a.Assemble()

m = BilinearForm(fes, symmetric=False)
m +=SymbolicBFI(CoefficientFunction(list(region_densities * region_hcs))* u*v)
m.Assemble()

f = LinearForm(fes)
f += SymbolicLFI(sources*v)

f.Assemble()

#%% time stepping

#Implicit Euler scheme
dt = 1
t_end = 100 #[s]
gfu.Set(initial_value) #set intial condition

Draw(gfu)

mstar = m.mat.CreateMatrix()
mstar.AsVector().data = m.mat.AsVector() + dt * a.mat.AsVector()
invmstar = mstar.Inverse(freedofs=fes.FreeDofs(), inverse="sparsecholesky")
res = gfu.vec.CreateVector()
 
t = 0.0
snapshots = []

# set Dirichlet values, keep internal
gfuDir = GridFunction(fes)
gfuDir.Set(dirichlet_temperature_const-gfu, BND)
gfu.vec.data += gfuDir.vec

while t < t_end:
    
    # res.data = m.mat * gfu.vec + dt * f.vec
    # gfu.Set(dirichlet_temperature_const, BND)
    # res.data += -mstar * gfu.vec  #homogenise the boundary condition

    res.data = f.vec - a.mat*gfu.vec
    gfu.vec.data += dt*invmstar * res
    t=round(t+dt,6) #round due to machine precision epsilon
    print('Time: ' + str(t))
    Redraw(blocking=True)
    
    if (t==5*dt):
        snap = np.zeros(len(gfu.vec.FV().NumPy()[:]))
        snap[:] = gfu.vec.FV().NumPy()[:]
        snapshots.append(snap)
        
solution_vector = gfu.vec.FV().NumPy()[:]    

##take a look at snapshots
#test = GridFunction(fes)
#Draw(test)

#for i in range(0,len(snapshots)):
#    test.vec.FV().NumPy()[:] = snapshots[i]
#    Redraw(blocking=True)
#    time.sleep(0.5)
 

